// 2019-08-02 ffgl@subpixels.com
// - Make inline shims for Compile(), FindUniform(), Set() and BindTexture() with const string& arguments
// - Add variants of Set() with uniformLocation arguments
// - Replace CompileVertexShader(), CompileGeometryShader(), CompileFragmentShader() by generic CompileShader()

#pragma once
#include "ffgl/FFGL.h"//For OpenGL
#include <vector>
#include <string>
#include <map>
#include <string>

namespace ffglex
{
/**
 * The Shader is a helper class you can use to compile glsl shader strings
 * and link them together into a shader program. You can then get the program id
 * and use it to bind the resulting shader to use it for the next rendering operation.
 */
class Shader final
{
public:
	Shader(); //The default constructor just initializes this object to represent no shader.
	~Shader();//Destructing this object just checks and lets you know if you've properly released the GL resources.

	void AddTransformFeedbackVarying( const std::string& varyingName );                              //Add a varying that will be captured using a transform feedback.
	bool Compile( const std::string& vertexShader, const std::string& geometryShader, const std::string& fragmentShader );
	inline bool Compile( const std::string& vertexShader, const std::string& fragmentShader )
	{
		return Compile( vertexShader, {}, fragmentShader );
	}

	void FreeGLResources();//Frees any GL resources that this shader is holding.

	bool IsReady() const;                       //Whether or not previous complilation succeeded and this shader is ready to be used for rendering.
	void Use() const;                           //Call glUseProgram with the current programID
	GLuint GetGLID() const;                     //Gets the OpenGL ID that represents the linked shader program.
	GLint FindUniform( const std::string& name ) const;//Finds the uniform location that needs to be used to addess a uniform with a certain name.

	void Set( const std::string& name, float value );
	void Set( const std::string& name, float v1, float v2 );
	void Set( const std::string& name, float v1, float v2, float v3 );
	void Set( const std::string& name, float v1, float v2, float v3, float v4 );
	void Set( const std::string& name, int value );

	void Set( GLint uniformLocation, float value );
	void Set( GLint uniformLocation, float v1, float v2 );
	void Set( GLint uniformLocation, float v1, float v2, float v3 );
	void Set( GLint uniformLocation, float v1, float v2, float v3, float v4 );
	void Set( GLint uniformLocation, int value );

	void BindTexture( const std::string& name, int texture, const ffgl::TextureStruct& fbo );

	Shader& operator=( const Shader& ) = delete;
	Shader& operator=( Shader&& ) = delete;

private:
	//Internal utility function that tries to compile a shader. Doesn't clean up on failure.
	bool CompileShader( const GLuint shaderID, const std::string& shaderCode, const std::string& shaderName );

	//Internal utility function that tries to link the two previously compiled vertex and fragment shaders into a program. Doesn't clean up on failure.
	bool LinkProgram();

	GLuint vertexShaderID;  //!< The ID OpenGL gave our vertex shader. 0 for invalid.
	GLuint geometryShaderID;//!< The ID OpenGL gave our geometry shader. 0 if not present or invalid.
	GLuint fragmentShaderID;//!< The ID OpenGL gave our fragment shader. 0 for invalid.
	GLuint programID;       //!< The ID OpenGL gave our shader program. BindTexture this to use this shader. 0 for invalid.

	std::vector< std::string > transformFeedbackVaryings;//!< The varyings that will be captured using a transform feedback. Ordered in the order of capturing.
};

}//End namespace ffglex